"""
This module implements classes that represents Certification Authorities (CA).
"""
import shutil

import re

import json
import os
import python_freeipa
from cryptography import x509
from hashlib import md5
from pathlib import Path, PosixPath
from python_freeipa import exceptions
from python_freeipa.client_meta import ClientMeta
from shutil import rmtree, copy2
from socket import gethostname

from SCAutolib import TEMPLATES_DIR, logger, run, LIB_DIR, LIB_DUMP_CAS, \
    LIB_BACKUP
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.models.file import OpensslCnf
from SCAutolib.enums import CAType


class BaseCA:
    dump_file: Path = None
    ca_type: str = None
    _ca_cert: Path = None
    _ca_key: Path = None
    _ca_pki_db: Path = Path("/etc/sssd/pki/sssd_auth_ca_db.pem")
    _ca_original_path: Path = LIB_BACKUP.joinpath("ca-db-original.backup")

    @property
    def cert(self):
        return self._ca_cert

    def request_cert(self, csr, username: str, cert_out: Path):
        """
        Request certificate from CA for given username

        :param csr: path to CSR
        :type csr: str
        :param username: subject for the certificate
        :type username: str
        :param cert_out: path where the certificate should be duplicated.
            Default None
        :type cert_out: pathlib.Path
        :return: None or path where the certificate is stored.
        :rtype: None or pathlib.Path
        """
        ...

    def setup(self):
        """
        Configure the CA
        """
        ...

    def update_ca_db(self):
        """
        Update /etc/sssd/pki/sssd_auth_ca_db.pem with certificate defined in CA
        object.
        """
        with self._ca_cert.open("r") as f_cert:
            root_cert = f_cert.read()

        if self._ca_pki_db.exists():
            with self._ca_pki_db.open() as f:
                with self._ca_original_path.open('w') as backup:
                    backup.write(f.read())
            # Check if current CA cert is already present in the sssd auth db
            with self._ca_pki_db.open("a+") as f:
                f.seek(0)
                data = f.read()
                if root_cert not in data:
                    f.write(root_cert)
        else:
            # Create /etc/sssd/pki directory if it doesn't exist
            self._ca_pki_db.parents[0].mkdir(exist_ok=True)
            with self._ca_pki_db.open("w") as f:
                f.write(root_cert)
        logger.debug(
            f"CA certificate {self._ca_cert} is copied to {self._ca_pki_db}")
        # Restoring SELinux context on the sssd auth db
        run(f"restorecon -v {self._ca_pki_db}")
        logger.info("Local CA is updated")

    def restore_ca_db(self):
        """
        restores /etc/sssd/pki/sssd_auth_ca_db.pem to the state it was before.
        """
        if self._ca_original_path.exists():
            logger.debug("Found original version of sssd_auth_ca_db.pem")
            with self._ca_original_path.open() as backup, \
                    self._ca_pki_db.open("w") as f:
                f.write(backup.read())
            self._ca_original_path.unlink()
        else:
            logger.debug("Original version of sssd_auth_ca_db.pem not found")
            if self._ca_pki_db.exists():
                self._ca_pki_db.unlink()
        logger.info(f"Restored {self._ca_pki_db} to original version")

    def sign_cert(self):
        """
        Sign the certificate
        """
        ...

    def revoke_cert(self, cert: Path):
        """
        Revoke the certificate

        :param cert:
        :type cert: pathlib.Path
        """
        ...

    @staticmethod
    def load(json_file):
        """
        Load CA from JSON file.
        :return: CA object
        """
        with json_file.open("r") as f:
            cnt = json.load(f)

        if cnt["ca_type"] == CAType.ipa:
            ca = IPAServerCA(ip_addr=cnt["_ipa_server_ip"],
                             server_hostname=cnt["_ipa_server_hostname"],
                             root_passwd=cnt["_ipa_server_root_passwd"],
                             admin_passwd=cnt["_ipa_server_admin_passwd"],
                             client_hostname=cnt["_ipa_client_hostname"],
                             domain=cnt["_ipa_server_domain"],
                             realm=cnt["_ipa_server_realm"])
        elif cnt["ca_type"] == CAType.custom:
            ca = CustomCA(cnt)
        elif cnt["ca_type"] == CAType.local:
            ca = LocalCA(root_dir=cnt["root_dir"])
        else:
            raise SCAutolibException("CA object has unknown type. Only ipa, "
                                     "custom and local types are supported. CA "
                                     "object not loaded")

        logger.debug(f"CA {cnt['name']} is loaded from file {json_file}")
        return ca


class LocalCA(BaseCA):
    """
    Represents local CA that is created as CA for virtual cards.
    """
    template = Path(TEMPLATES_DIR, "ca.cnf")
    ca_type = CAType.local
    ca_name = "local_ca"
    dump_file = LIB_DUMP_CAS.joinpath(f"{ca_name}.json")

    def __init__(self, root_dir: Path = None, cnf: OpensslCnf = None):
        """
        Class for local CA. Initialize required attributes, real setup is made
        by LocalCA.setup() method

        :param root_dir: Path to root directory of the CA. By default, is in
            /etc/SCAutolib/ca
        :type: Path
        :param cnf: object representing openssl cnf file
        :type cnf: OpensslCnf
        """
        self.name = LocalCA.ca_name
        self.ca_type = LocalCA.ca_type
        self.root_dir: Path = Path("/etc/SCAutolib/ca") if root_dir is None \
            else Path(root_dir)
        if not self.root_dir.exists():
            raise FileNotFoundError("Root directory of CA does not exist.")
        self._conf_dir: Path = self.root_dir.joinpath("conf")
        self._newcerts: Path = self.root_dir.joinpath("newcerts")
        self._certs: Path = self.root_dir.joinpath("certs")
        self._crl: Path = self.root_dir.joinpath("crl", "root.pem")
        self._ca_pki_db = BaseCA._ca_pki_db

        self._ca_cnf: OpensslCnf = cnf if cnf else OpensslCnf(
            conf_type="CA",
            filepath=self.root_dir.joinpath("ca.cnf"),
            replace=str(self.root_dir))
        self._ca_cert = self.root_dir.joinpath("rootCA.pem")
        self._ca_key = self.root_dir.joinpath("rootCA.key")

        self._serial: Path = self.root_dir.joinpath("serial")
        self._index: Path = self.root_dir.joinpath("index.txt")

    @property
    def cnf(self):
        return self._ca_cnf

    @cnf.setter
    def cnf(self, cnf: OpensslCnf):
        if not cnf.path.exists():
            raise SCAutolibException("CNF file does not exist")
        self._ca_cnf = cnf

    def to_dict(self):
        """
        Customising default property for better serialisation for storing to
        JSON format.

        :return: dictionary with all values. Path objects are typed to string.
        :rtype: dict
        """
        dict_ = {k: str(v) if type(v) is PosixPath else v
                 for k, v in super().__dict__.items()}
        if self._ca_cnf:
            dict_["_ca_cnf"] = str(self._ca_cnf.path)
        return dict_

    @property
    def is_installed(self):
        """
        Check if the local CA is installed
        """
        try:
            result = all([self.root_dir.exists(),
                          self._ca_cert.exists(),
                          self._ca_key.exists(),
                          self._ca_cnf.path.exists(),
                          self._conf_dir.exists(),
                          self._newcerts.exists(),
                          self._certs.exists(),
                          self._crl.exists(),
                          self._serial.exists(),
                          self._index.exists()])
            if result and self._ca_pki_db.exists():
                with self._ca_pki_db.open("r") as f:
                    with self._ca_cert.open("r") as cert:
                        result &= cert.read() in f.read()
        except Exception as e:
            logger.error(e)
            return False
        return result

    def setup(self):
        """
        Creates directory and file structure needed by local CA.
        """
        if self._ca_cnf is None:
            raise SCAutolibException("CA CNF file is not set")
        elif not self._ca_cnf.path.exists():
            raise SCAutolibException("CA CNF does not exist")

        self.root_dir.mkdir(parents=True, exist_ok=True)
        self._newcerts.mkdir(exist_ok=True)
        self._certs.mkdir(exist_ok=True)
        self._crl.parent.mkdir(exist_ok=True)

        with self._serial.open("w") as f:
            f.write("01")

        self._index.touch()

        # Generate self-signed certificate
        cmd = ['openssl', 'req', '-batch', '-config', self._ca_cnf.path,
               '-x509', '-new', '-nodes', '-newkey', 'rsa:2048', '-keyout',
               self._ca_key, '-sha256', '-set_serial', '0',
               '-extensions', 'v3_ca', '-out', self._ca_cert]
        run(cmd, check=True)
        logger.info(f"CA private key is generated into {self._ca_key}")
        logger.info(
            f"CA self-signed certificate is generated into {self._ca_cert}")

        # Configuring CRL
        run(['openssl', 'ca', '-config', self._ca_cnf.path, '-gencrl',
             '-out', self._crl], check=True)

        logger.info("Local CA files are prepared")

    def request_cert(self, csr: Path, username: str,
                     cert_out: Path = None) -> Path:
        """
        Create the certificate from CSR and sign it. Certificate is stored
        in the <root ca directory>/ca/newcerts directory with name username.pem

        :param csr: path to CSR
        :type csr: pathlib.Path
        :param username: subject in the CSR
        :type username: str
        :param cert_out: path where the certificate should be duplicated.
                         Can be a directory or a file. If a file, .pem extension
                         would be set to the filename. If not specified,
                         certificate would be created in default directory and
                         filename  <root ca directory>/certs/<username>.pem
        :type cert_out: pathlib.Path
        :return: returns path to the signed certificate
        :rtype: pathlib.Path
        """
        if cert_out is not None:
            if cert_out.is_dir():
                cert_out = cert_out.joinpath(f"{username}.pem")
            elif cert_out.is_file() and cert_out.suffixes[-1] != ".pem":
                cert_out = cert_out.with_suffix(".pem")
        else:
            cert_out = self._certs.joinpath(f"{username}.pem")
        cmd = ["openssl", "ca", "-config", self._ca_cnf.path,
               "-batch", "-keyfile", str(self._ca_key), "-in", str(csr),
               "-notext", "-days", "365", "-extensions", "usr_cert",
               "-out", str(cert_out)]
        run(cmd, check=True)
        return cert_out

    def revoke_cert(self, cert: Path):
        """
        Revoke given certificate

        :param cert: path to the certificate
        :type cert: pathlib.Path
        """
        cmd = ['openssl', 'ca', '-config', self._ca_cnf.path, '-revoke', cert]
        run(cmd, check=True)
        cmd = ['openssl', 'ca', '-config', self._ca_cnf.path, '-gencrl',
               '-out', self._crl]
        run(cmd, check=True)
        logger.info("Certificate is revoked")

    def cleanup(self):
        """
        Remove the root directory of the CA
        """
        logger.warning(f"Removing local CA {self.root_dir}")
        for file in self.root_dir.iterdir():
            if file.is_file():
                file.unlink()
            elif file.is_dir():
                shutil.rmtree(file)

        if self.dump_file.exists():
            self.dump_file.unlink()
            logger.debug(f"Removed {self.dump_file} dump file")

        logger.info(f"Local CA from {self.root_dir} is removed")


class CustomCA(BaseCA):
    """
    :TODO: CustomCA is not tested yet and it's not functional until physical
        cards testing with removinator is implemented
    Represents CA for physical cards. Physical cards are often read-only and
    rootCA certs or bundles are provided with a card. This class provides
    methods for manipulation with rootCA certs of physical cards.
    """
    ca_type = CAType.custom

    def __init__(self, card: dict):
        """
        Initialize required attributes
        """
        self.ca_type = CustomCA.ca_type
        self.name = card["ca_name"]
        self.ca_cert = card["ca_cert"]
        self.dump_file = LIB_DUMP_CAS.joinpath(f"{self.name}.json")
        self.root_dir: Path = LIB_DIR.joinpath(self.name)
        self._ca_cert = self.root_dir.joinpath(f"{self.name}.pem")
        self._ca_pki_db: Path = BaseCA._ca_pki_db

    def setup(self):
        """
        Create rootCA file. Actually, copy cert from conf.json
        """
        self.root_dir.mkdir(parents=True, exist_ok=True)
        if self.ca_cert is None:
            raise SCAutolibException(
                f"CA cerf for {self.name} not found")
        with self._ca_cert.open('w') as newcert:
            newcert.write(self.ca_cert)
        logger.info("Local CA files are prepared")

    def to_dict(self):
        """
        Customising default property for better serialisation for storing to
        JSON format.

        :return: dictionary with all values. Path objects are typed to string.
        :rtype: dict
        """
        dict_ = {k: str(v) if type(v) is PosixPath else v
                 for k, v in super().__dict__.items()}
        return dict_


class IPAServerCA(BaseCA):
    """
    Class represents IPA server with integrated CA. Through this class
    communication with IPA server is made primarily using
    ``python_freeipa.client_meta.ClientMeta`` class.

    During setup of the IPA client
    on the system, script generated on IPA server side has to be executed for
    setting up the client for smart card authentication. For this reason SSH
    connection is made to the server and the script is fetched in frame of
    ``IPAServerCA.create()`` method.
    """
    ca_type = CAType.ipa
    ca_name = "IPA"
    _ca_cert: Path = Path("/etc/ipa/ca.crt")
    _ipa_server_ip: str = None
    _ipa_server_hostname: str = None
    _ipa_server_domain: str = None
    _ipa_server_admin_passwd: str = None
    _ipa_server_realm: str = None
    _ipa_client_hostname: str = None
    _ipa_server_root_passwd: str = None
    _ipa_client_script = Path(LIB_DIR, "ipa-client-sc.sh")
    meta_client: ClientMeta = None
    dump_file = LIB_DUMP_CAS.joinpath("ipa-server.json")

    def __init__(self, ip_addr: str, server_hostname: str, domain: str,
                 admin_passwd: str, root_passwd: str, client_hostname: str,
                 realm: str = None):
        """
        Initialize object for IPA client for given IPA server. Also, creates
        meta client (python_freeipa.client_meta.ClientMeta) logged in to the
        server and ready-to-use.

        :param ip_addr: IP address of the IPA server
        :type ip_addr: str
        :param server_hostname: Hostname of the IPA server
        :type server_hostname: str
        :param domain: Domain name of the IPA server
        :type domain: str
        :param admin_passwd: Password for admin user on the IPA server
        :type admin_passwd: str
        :param root_passwd: Password for root user on the IPA server
                            (system user)
        :type root_passwd: str
        :param client_hostname: Hostname for the client. This name would be set
                                on the client host
        :type client_hostname: str
        :param realm: Kerberos realm. If not set, domain in upper cases would
                      be used instead
        :type realm: str
        """
        self.ca_type = IPAServerCA.ca_type
        self.name = IPAServerCA.ca_name
        self._ipa_server_ip = ip_addr
        self._ipa_server_hostname = server_hostname
        self._add_to_hosts()  # So we can log in to the IPA before setup

        self._ipa_server_domain = domain
        self._ipa_server_admin_passwd = admin_passwd
        self._ipa_server_realm = realm if realm is not None else domain.upper()
        self._ipa_client_hostname = client_hostname
        self._ipa_server_root_passwd = root_passwd

        self._meta_client_login()

    @property
    def is_installed(self):
        """
        :return: True, if IPA client is installed on the system (/etc/ipa
            directory contains ca.crt file from IPA server), otherwise False
        :rtype: bool
        """
        d = Path("/etc/ipa")
        result = d.exists()
        if result:
            result = d.joinpath("ca.crt")
        return result

    @property
    def domain(self):
        return self._ipa_server_domain

    def to_dict(self):
        """
        Customising default property for better serialisation for storing to
        JSON format.

        :return: dictionary with all values. Path objects are typed to string.
        :rtype: dict
        """
        dict_: dict = super().__dict__.copy()
        dict_.pop("meta_client")
        return dict_

    def setup(self):
        """
        Setup IPA client for IPA server. After IPA client is installed, system
        would be configured for smart card login with IPA using script from
        IPA server obtained via SSH.
        """
        logger.info(f"Start setup of IPA client on the system for "
                    f"{self._ipa_server_hostname} IPA server.")

        self._add_to_resolv()
        self._set_hostname()

        logger.info("Installing IPA client")
        try:
            run(["ipa-client-install", "-p", "admin",
                 "--password", self._ipa_server_admin_passwd,
                 "--server", self._ipa_server_hostname,
                 "--domain", self._ipa_server_domain,
                 "--realm", self._ipa_server_realm,
                 "--hostname", self._ipa_client_hostname,
                 "--force", "--force-join", "--no-ntp",
                 "--no-dns-sshfp", "--mkhomedir", "--unattended"],
                input="yes")
        except:
            logger.critical("Installation of IPA client is failed")
            rmtree("/etc/ipa/*")
            logger.debug("Directory /etc/ipa is removed")
            raise
        logger.debug("IPA client is installed")

        try:
            copy2("/tmp/cont-data/config-client-for-smart-card-auth.sh",
                  self._ipa_client_script)
            logger.info("Script for setting up IPA client for smart cards was "
                        f"found and copied to {self._ipa_client_script}")
        except FileNotFoundError:
            logger.info("Script for setting up IPA client for smart cards was "
                        "not found. It will be generated on IPA server and "
                        "fetched")
            self._get_sc_setup_script()
        run("kinit admin", input=self._ipa_server_admin_passwd)
        run(f'bash {self._ipa_client_script} /etc/ipa/ca.crt', check=True)
        logger.debug("Setup of IPA client for smart card is finished")

        self._meta_client_login()

        policy = self.meta_client.pwpolicy_show(a_cn="global_policy")["result"]
        if ["0"] != policy["krbminpwdlife"]:
            self.meta_client.pwpolicy_mod(a_cn="global_policy",
                                          o_krbminpwdlife=0)
            logger.debug("Minimal kerberos password lifetime is set to 0 days")
        if "365" not in policy["krbmaxpwdlife"]:
            self.meta_client.pwpolicy_mod(a_cn="global_policy",
                                          o_krbmaxpwdlife=365)
            logger.debug(
                "Maximum kerberos password lifetime is set to 365 days")

        # TODO: add to restore client host name
        logger.info("IPA client is configured on the system.")

    def _meta_client_login(self):
        """
        Login to admin user via IPA meta client.
        """
        try:
            self.meta_client: ClientMeta = ClientMeta(self._ipa_server_hostname,
                                                      verify_ssl=False)
            self.meta_client.login("admin", self._ipa_server_admin_passwd)
            logger.info("Connected to IPA via meta client")
        except python_freeipa.exceptions.BadRequest:
            logger.warning("Can't login to the IPA server. "
                           "Client might be not configured")

    def _set_hostname(self):
        """
        Set hostname for specified IPA client hostname
        """
        run(f"hostnamectl set-hostname {self._ipa_client_hostname} --static")
        logger.debug(f"Hostname is set to {self._ipa_client_hostname}")

    def _add_to_resolv(self):
        """
        Add new nameserver (IPA) to /etc/resolv.conf and lock this file for
        editing
        """
        nameserver = f"nameserver {self._ipa_server_ip}"
        pattern = rf"^nameserver\s+{self._ipa_server_ip}\s*"
        with open("/etc/resolv.conf", "r") as f:
            cnt = f.read()
        logger.debug(f"Original resolv.conf:\n{cnt}")
        if re.match(pattern, cnt) is None:
            logger.warning(f"Nameserver {self._ipa_server_ip} is not "
                           "present in /etc/resolv.conf. Adding...")
            cnt = (nameserver + "\n" + cnt)
            with open("/etc/resolv.conf", "w") as f:
                f.write(cnt)
            logger.info(
                "IPA server is added to /etc/resolv.conf as first nameserver")
            run("chattr -i /etc/resolv.conf")
            logger.info("File /etc/resolv.conf is blocked for editing")

        with open("/etc/resolv.conf", "r") as f:
            logger.debug(f"New resolv.conf\n{f.read()}")

    def _add_to_hosts(self):
        """
        Add IPA server (IP address and hostname) to /etc/hosts
        """
        entry = f"{self._ipa_server_ip} {self._ipa_server_hostname}"
        with open("/etc/hosts", "r+") as f:
            cnt = f.read()
            if entry not in cnt:
                f.write(f"\n{entry}\n")
                logger.warning(
                    f"New entry {entry} for IPA server is added to /etc/hosts")
            logger.info(
                f"Entry for IPA server {entry} presents in the /etc/hosts")

    def _get_sc_setup_script(self):
        """
        Fetch script for smart card setup of IPA client and place it to
        predefined location. Script is generated only on IPA server.
        Fetching is done by connecting to the host via SSH.
        """
        import paramiko
        from invoke import Responder
        from fabric.connection import Connection

        class __PKeyChild(paramiko.PKey):
            """This child class is need to fix SSH connection with MD5 algorithm
            in FIPS mode

            This is just workaround until PR in paramiko would be accepted
            https://github.com/paramiko/paramiko/issues/396. After this PR is
            merged, delete this class
            """

            def get_fingerprint_improved(self):
                return md5(self.asbytes(), usedforsecurity=False).digest()

        kinitpass = Responder(
            pattern=f"Password for admin@{self._ipa_server_realm}: ",
            response=f"{self._ipa_server_admin_passwd}\n")
        logger.debug("Start receiving client script for setting up smart card "
                     "on IPA client")
        with Connection(self._ipa_server_ip, user="root",
                        connect_kwargs={
                            "password": self._ipa_server_root_passwd}) as c:
            # Delete this block when PR in paramiko will be accepted
            # https://github.com/paramiko/paramiko/issues/396
            #### noqa:E266
            paramiko.PKey.get_fingerprint = \
                self.__PKeyChild.get_fingerprint_improved
            c.client = paramiko.SSHClient()
            c.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            #### noqa:E266
            c.open()
            # in_stream = False is required because while testing with pytest
            # it collision appears with capturing of the output.
            logger.debug("Running kinit on the IPA server")
            c.run("kinit admin", pty=True,
                  watchers=[kinitpass], in_stream=False)
            result = c.run("ipa-advise config-client-for-smart-card-auth",
                           hide=True, in_stream=False)
            logger.debug("Script is generated on server side")
            with open(self._ipa_client_script, "w") as f:
                f.write(result.stdout)

        if os.stat(self._ipa_client_script).st_size == 0:
            msg = "Script for IPA client smart card setup is not correctly " \
                  "copied to the host"
            logger.error(result.stdout)
            logger.error(result.stderr)
            raise SCAutolibException(msg)

        logger.debug("File for setting up IPA client for smart cards is "
                     f"copied to {self._ipa_client_script}")

    def request_cert(self, csr: Path, username: str, cert_out: Path):
        """
        Request certificate using CSR from IPA CA for given username. It is
        a wrapper around the python_freeipa.client_meta.ClientMeta.cert_request
        method. It works with a file, extracts CSR data from it and then
        stores in PEM format adding required prefix and suffix as in normal
        certificate and. If cert_out is a directory, then certificate would be
        stored in this directory with name <username>.pem. If it is a file,
        then check if it has PEM extension. If not, append .pem suffix to the
        name.

        :param csr: path to CSR
        :type csr: patlib.Path
        :param username: subject for the certificate
        :type username: str
        :param cert_out: path where the certificate is stored. Can be a
                         directory or a file.
        :type cert_out: patlib.Path

        :return: Path to the PEM certificate.
        :rtype: patlib.Path
        """
        with csr.open() as f:
            csr_content = f.read()
        r = self.meta_client.cert_request(a_csr=csr_content,
                                          o_principal=username)
        logger.debug(r)
        cert = r["result"]["certificate"]

        if cert_out.is_dir():
            cert_out = cert_out.joinpath(f"{username}.pem")
        else:
            cert_out = cert_out.with_suffix(".pem")

        with cert_out.open("w") as f:
            f.write("-----BEGIN CERTIFICATE-----\n"
                    f"{cert}\n"
                    f"-----END CERTIFICATE-----")
        return cert_out

    def add_user(self, user):
        """
        Add given user to IPA server. It is a wrapper on the
        python_freeipa.client_meta.ClientMeta.user_add method. Just extracts
        necessary fields from IPAUser object and pass them to the method. As a
        result, o_givenname == o_uid == o_sn == o_cn for simplicity.

        :param user: User to be added to the IPA server.
        """
        r = self.meta_client.user_add(user.username, user.username,
                                      user.username, user.username,
                                      o_userpassword=user.password)
        logger.debug(r)
        logger.info(f"User {user.username} is added to the IPA server")

    def del_user(self, user):
        """
        Remove user from IPA server.
        :param user: User to be deleted

        :raise AssertionError: If operation is failed. Fail is detected in
            return value from request to IPA server.
        """
        r = self.meta_client.user_del(user.username)["result"]
        logger.debug(r)
        logger.info(f"User {user.username} is removed from the IPA server")

    def revoke_cert(self, cert_path: Path):
        """
        Revoke given certificate on the IPA server. It is a wrapper on the
        python_freeipa.client_meta.ClientMeta.revoke_cert method. It extracts
        serial number of the certificate from the file

        :param cert_path: Path to the certificate in PEM format

        """
        with cert_path.open("rb") as f:
            cert = x509.load_pem_x509_certificate(f.read())
        r = self.meta_client.cert_revoke(cert.serial_number)
        logger.debug(r)
        logger.info(f"Certificate {cert.serial_number} is revoked")
        return cert.serial_number

    def cleanup(self):
        """
        Remove IPA client from the system and from the IPA server

        :raise subprocess.CalledProcessError: by run function
        """

        logger.warning("Removing IPA client from the host "
                       f"{gethostname()}")
        try:
            r = self.meta_client.host_del(
                a_fqdn=gethostname(), o_updatedns=True)["result"]
            logger.debug(r)
            assert r["failed"] == [], "Deleting of the host is failed"
        except exceptions.NotFound:
            logger.error(f"Current hostname ({gethostname()}) is not found "
                         f"on the IPA server")
        # Return code 2 means that the IPA client is not configured
        run(["ipa-client-install", "--uninstall", "-U"], return_code=[0, 2])
        logger.info("IPA client is removed.")

    @property
    def ipa_server_hostname(self):
        return self._ipa_server_hostname



from enum import Enum, auto


class OSVersion(int, Enum):
    """
    Enumeration for Linux versions. Used for more convenient checks.
    """
    Fedora = 1
    RHEL_8 = 2
    RHEL_9 = 3
    RHEL_10 = 4
    CentOS_8 = 5
    CentOS_9 = 6
    CentOS_10 = 7


class CardType(str, Enum):
    physical = "physical"
    virtual = "virtual"


class UserType(str, Enum):
    local = "local"
    ipa = "ipa"


class CAType(str, Enum):
    local = "local"
    custom = "custom"
    ipa = "IPA"


class ReturnCode(Enum):
    """
    Enum for return codes
    """
    SUCCESS = 0
    MISSING_CA = auto()
    FAILURE = auto()
    ERROR = auto()
    EXCEPTION = auto()
    UNKNOWN = auto()



import json
import os
from pathlib import Path
from schema import Schema, Use
from shutil import rmtree
from typing import Union

from SCAutolib import exceptions, schema_cas, schema_user, schema_card
from SCAutolib import (logger, run, LIB_DIR, LIB_BACKUP, LIB_DUMP,
                       LIB_DUMP_USERS, LIB_DUMP_CAS, LIB_DUMP_CARDS,
                       LIB_DUMP_CONFS, TEMPLATES_DIR)
from SCAutolib.models import CA, file, user, card, authselect as auth
from SCAutolib.models.file import File, OpensslCnf
from SCAutolib.models.CA import BaseCA
from SCAutolib.enums import (CardType, UserType)
from SCAutolib.utils import (_check_selinux, _gen_private_key,
                             _install_packages, _check_packages,
                             dump_to_json, ca_factory)
from SCAutolib.isDistro import isDistro


class Controller:
    authselect: auth.Authselect = auth.Authselect()
    sssd_conf: file.SSSDConf = file.SSSDConf()
    lib_conf: dict = None
    _lib_conf_path: Path = None
    local_ca: CA.LocalCA = None
    ipa_ca: CA.IPAServerCA = None
    users: [user.User] = None
    dconf_file = File(filepath='/etc/dconf/db/local.d/gnome_disable_welcome',
                      template=Path(TEMPLATES_DIR, 'gnome_disable_welcome'))

    @property
    def conf_path(self):
        return self._lib_conf_path

    def __init__(self, config: Union[Path, str] = None, params: {} = None):
        """
        Constructor will parse and check input configuration file. If some
        required fields in the configuration are missing, CLI parameters
        would be checked if missing values are there. If not, an exception
        would be raised. After parsing the configuration file and filling
        internal values of the Controller object, other related objects (users,
        cards, CAs, Authselect, etc.) would be initialised, but any real action
        that would affect the system
        wouldn't be made.

        :param config: Path to configuration file with metadata for testing.
        :type config: pathlib.Path or str
        :param params: Parameters from CLI
        :type params: dict
        :return:
        """

        # Check params

        # Parse config file
        self.lib_conf = None
        if config:
            self._lib_conf_path = config.absolute() if isinstance(config, Path) \
                else Path(config).absolute()

            with self._lib_conf_path.open("r") as f:
                tmp_conf = json.load(f)
                if tmp_conf is None:
                    raise exceptions.SCAutolibException(
                        "Data are not loaded correctly.")
            self.lib_conf = self._validate_configuration(tmp_conf, params)

        self.users = []
        for d in (LIB_DIR, LIB_BACKUP, LIB_DUMP, LIB_DUMP_USERS, LIB_DUMP_CAS,
                  LIB_DUMP_CARDS, LIB_DUMP_CONFS):
            d.mkdir(exist_ok=True)

        if LIB_DUMP_CAS.joinpath("local_ca.json").exists():
            self.local_ca = BaseCA.load(LIB_DUMP_CAS.joinpath("local_ca.json"))

        if LIB_DUMP_CAS.joinpath("ipa-server.json").exists():
            self.ipa_ca = BaseCA.load(LIB_DUMP_CAS.joinpath("ipa-server.json"))

    def prepare(self, force: bool, gdm: bool, install_missing: bool,
                graphical: bool):
        """
        Prepare system for testing. This method provides complex configuration
        of system under test for testing including creation of CAs, users and
        smart cards in the system and objects that represents them in SCAutolib.
        Configuration is based on config file and CLI options.

        :param force: Defines if existing objects, files, users, services etc.
            should be erased or overwritten if they already exist. True stands
            for erase/overwrite. This parameter is forwarded to several methods
            and it can have slightly different meaning in each of them.
            For details see docstrings of the methods.
        :type force: bool

        :param gdm: If True, GDM package would be installed
        :type gdm: bool
        :param install_missing: If True, all missing packages would be
            installed
        :type install_missing: bool
        :param graphical: If True, GUI tests dependencies are installed
        :type graphical: bool
        """
        self.setup_system(install_missing, gdm, graphical)

        # Prepare CAs: Virtual cards are populated by certificates that are: a)
        # created locally and signed by local CA configured on the system under
        # test, or b) created and signed using FreeIPA.
        try:
            self.setup_local_ca(force=force)
        except exceptions.SCAutolibWrongConfig as e:
            logger.info(e)
        try:
            self.setup_ipa_client(force=force)
        except exceptions.SCAutolibWrongConfig as e:
            logger.info(e)

        for usr in self.lib_conf["users"]:
            self.setup_user(usr, force=force)

        # Create cards defined in config. For physical cards only objects will
        # be created while for virtual cards tokens will be created and enrolled
        for token in self.lib_conf["cards"]:
            # prepare CA objects for physical cards
            if token["card_type"] == CardType.physical:
                self.setup_custom_ca(token)
                self.setup_card(token)
            elif token["card_type"] == CardType.virtual:
                c = self.setup_card(token)
                self.enroll_card(c)

    def setup_system(self, install_missing: bool, gdm: bool, graphical: bool):
        """
        Do general system setup meaning package installation based on
        specifications in the configuration file, SSSD configuration,
        configurations for virtual smart cards, etc.

        :param install_missing: If True, all missing packages would be
            installed
        :type install_missing: bool
        :param gdm: If True, GDM package would be installed
        :type gdm: bool
        :param graphical: If True, GUI tests dependencies are installed
        :type graphical: bool
        :return:
        """
        for d in (LIB_DIR, LIB_BACKUP, LIB_DUMP, LIB_DUMP_USERS, LIB_DUMP_CAS,
                  LIB_DUMP_CARDS):
            d.mkdir(exist_ok=True)

        packages = ["opensc", "httpd", "sssd", "sssd-tools", "gnutls-utils",
                    "openssl", "nss-tools"]

        # Prepare for virtual cards
        if any(c["card_type"] == CardType.virtual
                for c in self.lib_conf["cards"]):
            packages += ["pcsc-lite-ccid", "pcsc-lite", "virt_cacard",
                         "vpcd", "softhsm"]
            extra_args = ""
            if isDistro(['rhel', 'centos'], version='10'):
                # TODO: use better approach later
                extra_args = " centos-stream-10-x86_64"
            run("dnf -y copr enable jjelen/vsmartcard{0}".format(extra_args))

        # Add IPA packages if needed
        if any([u["user_type"] != UserType.local
                for u in self.lib_conf["users"]]):
            packages += self._general_steps_for_ipa()

        # Check for installed packages
        missing = _check_packages(packages)
        if install_missing and missing:
            _install_packages(missing)
        elif missing:
            msg = "Can't continue. Some packages are missing: " \
                  f"{', '.join(missing)}"
            logger.critical(msg)
            raise exceptions.SCAutolibException(msg)

        if graphical:
            self.setup_graphical(install_missing, gdm)

        if not isDistro('fedora'):
            run(['dnf', 'groupinstall', "Smart Card Support", '-y',
                 '--allowerasing'])
            logger.debug("Smart Card Support group in installed.")
        else:
            # Fedora requires rsyslog as well
            run(['dnf', 'install', 'opensc', 'pcsc-lite-ccid', 'rsyslog', '-y'])
            run(['systemctl', 'start', 'rsyslog'])

        self.sssd_conf.create()
        self.sssd_conf.save()
        self._general_steps_for_virtual_sc()

        base_user = user.User("base-user", "redhat")
        base_user.add_user()
        dump_to_json(base_user)
        dump_to_json(user.User(username="root",
                               password=self.lib_conf["root_passwd"]))

    def setup_graphical(self, install_missing: bool, gdm: bool):
        packages = ["gcc", "tesseract", "ffmpeg-free"]

        if gdm:
            packages.append("gdm")

        missing = _check_packages(packages)
        if install_missing and missing:
            _install_packages(missing)
        elif missing:
            msg = "Can't continue with graphical. Some packages are missing: " \
                  f"{', '.join(missing)}"
            logger.critical(msg)
            raise exceptions.SCAutolibException(msg)

        if not isDistro('fedora'):
            run(['dnf', 'groupinstall', 'Server with GUI', '-y',
                '--allowerasing'])
            run(['pip', 'install', 'python-uinput'])
        else:
            # Fedora doesn't have server with GUI group so installed gdm
            # manually and also python3-uinput should be installed from RPM
            run(['dnf', 'install', 'gdm', 'python3-uinput', '-y'])
        # disable subscription message
        run(['systemctl', '--global', 'mask',
            'org.gnome.SettingsDaemon.Subscription.target'])
        # disable welcome message
        if not self.dconf_file.exists():
            self.dconf_file.create()
            self.dconf_file.save()
            run('dconf update')

    def setup_local_ca(self, force: bool = False):
        """
        Setup local CA based on configuration from the configuration file. All
        necessary files for this operation (e.g. CNF file for self-signed root
        certificate) would be created along the way.

        :param force: If local CA already exists in given directory, specifies
            if it should be overwritten
        :type force: bool
        :raises: SCAutolib.exceptions.SCAutolibWrongConfig
        """

        if "local_ca" not in self.lib_conf["ca"]:
            msg = "Section for local CA is not found in the configuration file"
            raise exceptions.SCAutolibWrongConfig(msg)

        ca_dir: Path = self.lib_conf["ca"]["local_ca"]["dir"]
        ca_dir.mkdir(exist_ok=True, parents=True)

        cnf = OpensslCnf(ca_dir.joinpath("ca.cnf"), "CA", str(ca_dir))
        self.local_ca = ca_factory(path=ca_dir, cnf=cnf, create=True)
        if force:
            logger.warning(f"Removing previous local CA from {ca_dir}")
            self.local_ca.cleanup()
        cnf.create()
        cnf.save()
        self.local_ca.setup()
        self.local_ca.update_ca_db()
        run(["systemctl", "restart", "sssd"], sleep=5)

        logger.info(f"Local CA is configured in {ca_dir}")

        dump_to_json(self.local_ca)

    def setup_custom_ca(self, card_data: dict):
        if card_data["card_type"] == CardType.physical:
            ca = ca_factory(create=True, card_data=card_data)
            ca.setup()
            if not ca._ca_cert.is_file():
                raise FileNotFoundError(f"File not found: {ca._ca_cert}")
            dump_to_json(ca)

    def setup_ipa_client(self, force: bool = False):
        """
        Configure IPA client for given IPA server on current host. IPA server
        should be already up and running for correct configuration of the IPA
        client

        :param force: If IPA Client is already configured on the system,
            specifies if it should be removed before configuring a new client.
        :type force: bool
        :raises: SCAutolib.exceptions.SCAutolibWrongConfig
        """
        if "ipa" not in self.lib_conf["ca"]:
            msg = "Section for IPA is not found in the configuration file"
            raise exceptions.SCAutolibWrongConfig(msg)
        self.ipa_ca = CA.IPAServerCA(**self.lib_conf["ca"]["ipa"])

        if self.ipa_ca.is_installed:
            logger.warning("IPA client is already configured on this system.")
            if not force:
                logger.info("Set force argument to True if you want to remove "
                            "previous installation.")
                return
            self.ipa_ca.cleanup()
        else:
            logger.info("IPA client is not configured on the system")
        self.ipa_ca.setup()
        self.sssd_conf.update_default_content()
        self.sssd_conf.set(key="domains",
                           value=f"shadowutils, {self.ipa_ca.domain}",
                           section="sssd")
        dump_to_json(self.ipa_ca)

    def setup_user(self, user_dict: dict, force: bool = False):
        """
        Configure the user on the specified system (local machine/CA).

        :param force: specify if the user should be re-created with its
            card directory
        :type force: bool
        :param user_dict: set of values to initialise the user
        :type user_dict: dict
        :return: the user object
        """
        new_user = None

        if user_dict["user_type"] == UserType.local:
            new_user = user.User(username=user_dict["name"],
                                 password=user_dict["passwd"])
            if force:
                new_user.delete_user()
            new_user.add_user()

        else:
            if self.ipa_ca is None:
                msg = "Can't proceed in configuration of IPA user because no " \
                      "IPA Client is configured"
                raise exceptions.SCAutolibException(msg)
            new_user = user.IPAUser(ipa_server=self.ipa_ca,
                                    username=user_dict["name"],
                                    password=user_dict["passwd"])
            if force:
                new_user.delete_user()
            new_user.add_user()
        self.users.append(new_user)
        dump_to_json(new_user)
        return new_user

    def setup_card(self, card_dict: dict, force: bool = False):
        """
        Create card object. Card object should contain its root CA cert as it
        represents general card (i.e. including physical read-only cards).

        :param card_dict: Dictionary containing card attributes
        :type card_dict: dict
        :param force: If its true and card directory exists it will be removed
        :type force: bool
        """
        card_dir: Path = Path("/root/cards", card_dict["name"])
        card_dir.mkdir(parents=True, exist_ok=True)

        if force and card_dir.exists():
            rmtree(card_dir)

        if card_dict["card_type"] == CardType.physical:
            new_card = card.PhysicalCard(card_dict, card_dir=card_dir)
        elif card_dict["card_type"] == CardType.virtual:
            hsm_conf = self.prepare_softhsm_config(card_dir)
            new_card = card.VirtualCard(card_dict, softhsm2_conf=hsm_conf.path,
                                        card_dir=card_dir)
            # card needs to know some details of its user
            new_card.user = self.link_user_to_card(new_card)
            if new_card.user.user_type == UserType.local:
                new_card.cnf = self.prepare_user_cnf(new_card)
            if force:
                self.revoke_certs(new_card)
            new_card.create()
        else:
            raise NotImplementedError("Other card types than 'physical' and "
                                      "'virtual' are not supported")

        dump_to_json(new_card)
        return new_card

    def link_user_to_card(self, card: card.VirtualCard):
        for card_user in self.users:
            if card_user.username == card.cardholder:
                return card_user

    def prepare_softhsm_config(self, card_dir: Path = None):
        """Prepare SoftHSM2 config for virtual card"""
        filepath = card_dir.joinpath("sofhtsm2.conf")
        hsm_conf = file.SoftHSM2Conf(filepath, card_dir=card_dir)
        hsm_conf.create()
        hsm_conf.save()
        return hsm_conf

    def prepare_user_cnf(self, card: card.VirtualCard):
        """Prepare user openssl cnf"""
        cnf_path = card.card_dir.joinpath(f"{card.cardholder}.cnf")
        cnf = file.OpensslCnf(filepath=cnf_path, conf_type="user",
                              replace=[card.cardholder, card.CN])
        cnf.create()
        cnf.save()
        return cnf.path

    def revoke_certs(self, card: card.VirtualCard):
        if card.cert and card.cert.exists():
            if card.user.user_type == UserType.local:
                self.local_ca.revoke_cert(card.cert)
            else:
                self.ipa_ca.revoke_cert(card.cert)

    def enroll_card(self, card: card.VirtualCard):
        """
        Enroll the card - i.e. upload keys and certs to card. If private key
        and/or the certificate do not exist, new one's would be requested
        from corresponding CA.

        :param card: card object
        :type card: card.VirtualCard
        """
        logger.debug(f"Starting enrollment of the card {card.name}")
        if not card:
            raise exceptions.SCAutolibException(
                f"Card {card.name} is not initialized")

        if not card.key.exists():
            _gen_private_key(card.key)

        if not card.cert.exists():
            csr = card.gen_csr()
            ca = self.ipa_ca \
                if isinstance(card.user, user.IPAUser) else self.local_ca
            card.cert = ca.request_cert(csr, card.cardholder, card.cert)

        card.enroll()
        dump_to_json(card)

    def cleanup(self):
        """
        Clean the system after setup. This method restores the SSSD config file,
        deletes created users with cards, remove CA's (local and/or IPA Client)
        """
        users = {}

        for user_file in LIB_DUMP_USERS.iterdir():
            usr = user.User.load(user_file, ipa_server=self.ipa_ca)
            users[usr.username] = usr
            if usr.username != "root":
                usr.delete_user()

        for card_file in LIB_DUMP_CARDS.iterdir():
            if card_file.exists():
                card_obj = card.Card.load(card_file)
                if card_obj.card_type == CardType.virtual:
                    card_obj.user = users[card_obj.cardholder]
                    self.revoke_certs(card_obj)
                    card_obj.delete()

        if self.local_ca:
            self.local_ca.cleanup()
            self.local_ca.restore_ca_db()
        if self.ipa_ca:
            self.ipa_ca.cleanup()

        opensc_cache_dir = Path(os.path.expanduser('~') + "/.cache/opensc/")
        if opensc_cache_dir.exists():
            for cache_file in opensc_cache_dir.iterdir():
                cache_file.unlink()
        logger.debug("Removed opensc file cache")

        sssd_cache_dir = Path(os.path.expanduser('~sssd') + "/.cache/opensc/")
        if sssd_cache_dir.exists():
            for cache_file in sssd_cache_dir.iterdir():
                cache_file.unlink()
        logger.debug("Removed opensc file cache for sssd user")

        # file only created in graphical mode that is why it is removed.
        self.dconf_file.remove()

        self.sssd_conf.restore()
        pcscd_service = File("/usr/lib/systemd/system/pcscd.service")
        pcscd_service.restore()
        opensc_module = File("/usr/share/p11-kit/modules/opensc.module")
        opensc_module.restore()

    @staticmethod
    def _validate_configuration(conf: dict, params: {} = None) -> dict:
        """
        Validate schema of the configuration file. If some value is not present
        in the config file, this value would be looked in the CLI parameters

        :param conf: Configuration to be parsed (e.g. data loaded from
            JSON file)
        :type conf: dict
        :param params: CLI arguments
        :type params: dict
        :return: dictionary with parsed values from conf and params attributes.
            All values are retyped to specified type.
        :rtype: dict
        """
        # FIXME: any schema requires all values to be in the config file, and
        #  only IP address of IPA server is accepted from CLI arguments.
        #  Add loading of the values from params dict
        # IP regex
        # Specify validation schema for CAs

        # Specify general schema for whole config file
        schema = Schema({"root_passwd": Use(str),
                         "ca": schema_cas,
                         "users": [schema_user],
                         "cards": [schema_card]})

        return schema.validate(conf)

    @staticmethod
    def _general_steps_for_virtual_sc():
        """
        Prepare the system for virtual smart card. Preparation means to
        configure pcscd service and opensc module to work correctly
        with virtual smart card. Also, repository for installing virt_cacard
        package is added in this method.
        """

        _check_selinux()

        pcscd_service = File("/usr/lib/systemd/system/pcscd.service")
        pcscd_service.backup()
        exec_start = pcscd_service.get(section="Service", key="ExecStart")
        if "--auto-exit" in exec_start:
            exec_start = exec_start.replace("--auto-exit", "")
            pcscd_service.set(section="Service", key="ExecStart",
                              value=exec_start)
            pcscd_service.save()

        opensc_module = File("/usr/share/p11-kit/modules/opensc.module")
        opensc_module.backup()
        try:
            opensc_module.get("disable-in", separator=":")
        except exceptions.SCAutolibException:
            logger.warning("OpenSC module does not have option 'disable-in: "
                           "virt_cacard' set")
            opensc_module.set(key="disable-in", value="virt_cacard",
                              separator=": ")
            opensc_module.save()

        run(['systemctl', 'stop', 'pcscd.service', 'pcscd.socket', 'sssd'])
        rmtree("/var/lib/sss/mc/*", ignore_errors=True)
        rmtree("/var/lib/sss/db/*", ignore_errors=True)
        logger.debug(
            "Directories /var/lib/sss/mc/ and /var/lib/sss/db/ removed")

        run("systemctl daemon-reload")
        run("systemctl restart pcscd sssd")

        logger.debug("Copr repo for virt_cacard is enabled")

    @staticmethod
    def _general_steps_for_ipa():
        """
        General system preparation for installing IPA client on RHEL/Fedora

        :return: name of the IPA client package for current Linux
        """
        if isDistro(['rhel', 'centos'], version='8'):
            run("dnf module enable -y idm:DL1")
            run("dnf install @idm:DL1 -y")
            logger.debug("idm:DL1 module is installed")

        if isDistro('fedora'):
            return ["freeipa-client"]
        else:
            return ["ipa-client"]

    def get_user_dict(self, name):
        """
        Get user dictionary from the config file.

        :param name: name of the user
        :type name: str
        :return: user dictionary
        :rtype: dict
        """
        for user_dict in self.lib_conf["users"]:
            if user_dict["name"] == name:
                return user_dict
        raise exceptions.SCAutolibMissingUserConfig(name)

    def init_ca(self, local: bool = False):
        """
        Initialize CA.

        :param local: if True, local CA is initialized, otherwise IPA
        :type local: bool
        """
        if local:
            self.local_ca = CA.LocalCA(self.lib_conf["ca"]["local_ca"]["dir"])
            if not self.local_ca.cert.exists():
                raise exceptions.SCAutolibMissingCA(
                    f"CA certificate not found in {str(self.local_ca.cert)}")
        else:
            self.ipa_ca = CA.IPAServerCA(self.lib_conf["ca"]["ipa"])
            if not self.ipa_ca.is_installed:
                raise exceptions.SCAutolibMissingCA(
                    "IPA server CA is not installed")
        logger.info("CA is initialized")



import inspect
from time import sleep, time
from pathlib import Path

import cv2
import keyboard
import numpy as np
import pytesseract
import uinput
import logging

from SCAutolib import run, logger
from SCAutolib.isDistro import isDistro


class Screen:
    """Captures the screenshots."""

    def __init__(self, directory: str, html_file: str = None):
        """Init method
        :param directory: Path to directory, where the screenshots
            will be saved.
        """
        self.directory = directory
        self.html_file = html_file

        taken_images = [str(image).split('/')[-1]
                        for image in Path(directory).iterdir()]
        taken_images.sort(reverse=True)

        self.screenshot_num = 1
        if len(taken_images) > 0:
            self.screenshot_num = int(taken_images[0].split('.')[0]) + 1

    def screenshot(self, timeout: float = 30):
        """Runs ffmpeg to take a screenshot.

        :param timeout: Timeout in seconds. If ffmpeg cannot take screenshot
            before the specified timeout, an exception is raised.
        :return: Path to the screenshot
        :rtype: str
        """

        logger.debug(f"Taking screenshot number {self.screenshot_num}")

        filename = f'{self.directory}/{self.screenshot_num}.png'
        t_end = time() + timeout
        captured = False

        # if the ffmpeg command fails,
        # try screenshotting again until the timeout
        while time() < t_end and not captured:
            out = run(['ffmpeg', '-hide_banner', '-y', '-f',
                       'kmsgrab', '-i', '-', '-vf', 'hwdownload,format=bgr0',
                       '-frames', '1', '-update', '1',
                       filename], check=False, print_=False)

            if out.returncode == 0:
                captured = True
            else:
                logger.debug(f"ffmpeg failed with {out.returncode}. "
                             f"stdout: {out.stdout}, stderr: {out.stderr}")

        if not captured:
            raise Exception('Could not capture screenshot within timeout.')

        if self.html_file:
            with open(self.html_file, 'a') as fp:
                fp.write(
                    f"<img src=\"screenshots/{self.screenshot_num}.png\" "
                    f"alt=\"screenshot number {self.screenshot_num}\">"
                )

        self.screenshot_num += 1
        return filename


class Mouse:
    """Controls the mouse of the system under test"""

    def __init__(self):
        run(['modprobe', 'uinput'], check=True)

        # Maximum coordinate for both axis
        self.ABS_MAX = 2**16

        # initialize the uinput device
        self.device = uinput.Device((
            uinput.ABS_X + (0, self.ABS_MAX, 0, 0),
            uinput.ABS_Y + (0, self.ABS_MAX, 0, 0),
            uinput.BTN_LEFT,
            uinput.BTN_MIDDLE,
            uinput.BTN_RIGHT,
            uinput.REL_WHEEL,
        ))

        self.CLICK_HOLD_TIME = 0.1

    def move(self, x: float, y: float):
        """Moves the mouse cursor to specified absolute coordinate.
        Both coordinates are float numbers in range from 0 to 1.
        These get mapped to the screen resolution in the compositor.

        :param x: X coordinate of the cursor
        :param y: Y coordinate of the cursor
        """

        logger.info(f'Moving mouse to {x, y})')

        for uinput_axis, value in [(uinput.ABS_X, x), (uinput.ABS_Y, y)]:
            # Check if value between 0 and 1
            if not (0 <= value <= 1):
                raise ValueError("Values must be floats between 0 and 1")
            converted = int(value * self.ABS_MAX)
            self.device.emit(uinput_axis, converted, syn=False)

        # Both axes move at once
        self.device.syn()

    def click(self, button: str = 'left'):
        """Clicks any button of the mouse.

        :param button: mouse button to click, defaults to 'left'
            Possible values 'left', 'right' or 'middle'.
        """

        button_map = {
            'left': uinput.BTN_LEFT,
            'right': uinput.BTN_RIGHT,
            'middle': uinput.BTN_MIDDLE,
        }
        uinput_button = button_map[button]

        logger.info(f'Clicking the {button} mouse button')

        # press the button
        self.device.emit(uinput_button, 1)
        # wait a little
        sleep(self.CLICK_HOLD_TIME)
        # release the button
        self.device.emit(uinput_button, 0)


def image_to_data(path: str):
    """Convert screenshot into dataframe of words with their coordinates.

    :param path: path to the image to convert.
    """
    upscaling_factor = 2

    image = cv2.imread(path)
    grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    upscaled = cv2.resize(grayscale,
                          dsize=None,
                          fx=upscaling_factor,
                          fy=upscaling_factor,
                          interpolation=cv2.INTER_LANCZOS4)
    _, binary = cv2.threshold(upscaled, 120, 255, cv2.THRESH_BINARY_INV)
    df = pytesseract.image_to_data(binary, output_type='data.frame')

    yres, xres = binary.shape[:2]
    df[['left', 'width']] /= xres
    df[['top', 'height']] /= yres

    logger.debug(df)
    return df


def images_equal(path1: str, path2: str):
    """Compare two images, return True if they are completely identical.
    Images are considered identical, when their resolutions match and
    all pixel values are equal.
    Images can be in any format, that can be read using
    imread function from OpenCV.

    :param path1: Path to the first image.
    :param path2: Path to the second image.
    """
    im1 = cv2.imread(path1)
    im2 = cv2.imread(path2)

    # Is the resolution the same
    if im1.shape != im2.shape:
        return False

    # Check if value of every pixel is the same
    if np.bitwise_xor(im1, im2).any():
        return False

    return True


def action_decorator(func):
    """Decorator for all functions, that change the state of GUI.
    This decorator takes a screenshot before and after the action.
    The two screenshots are compared. If they are the same,
    an exception is raised.

    :param func: The function to be decorated
    """

    def wrapper(self,
                *args,
                wait_time=None,
                screenshot=True,
                check_difference=True,
                **kwargs):

        start_screenshot = self.screen.screenshot() if screenshot else None
        func(self, *args, **kwargs)
        sleep(wait_time or self.wait_time)
        end_screenshot = self.screen.screenshot() if screenshot else None

        if screenshot and check_difference:
            # If the checking is enabled
            # the action should change contents of the screen
            if images_equal(start_screenshot, end_screenshot):
                # If the screenshot before and after action are the same
                # an exception is raised
                raise Exception("Action did not change "
                                "the contents of the screen")

    return wrapper


def log_decorator(func):
    """Functions decorated with this will be logged when called.

    :param func: The function to be decorated
    """

    def wrapper(self, *args, **kwargs):
        # Format the arguments for logging
        kwargs_list = ["=".join((key, repr(value)))
                       for key, value in kwargs.items()]
        args_list = [repr(value) for value in list(args[1:])]
        all_args = ", ".join(args_list + kwargs_list)
        logger.info(f'Calling GUI.{func.__name__}({all_args})')
        func(self, *args, **kwargs)
    return wrapper


class GUI:
    """Represents the GUI and allows controlling the system under test."""

    def __init__(self, wait_time: float = 5, res_dir_name: str = None,
                 from_cli: bool = False):
        """Initializes the GUI of system under test.

        :param wait_time: Time to wait after each action
        :param custom_dir_name: Provide a custom name of the results dir under
        /tmp/SC-tests/. The default is `timestamp`_`caller func's name`.
        """

        self.wait_time = wait_time
        self.gdm_init_time = 10
        self.from_cli = from_cli
        # Create the directory for screenshots
        self.html_directory = Path("/tmp/SC-tests")
        if not self.html_directory.exists():
            self.html_directory.mkdir()
        if res_dir_name:
            self.html_directory = self.html_directory.joinpath(res_dir_name)
        elif from_cli:
            run_dirs = [str(run_dir).split('/')[-1]
                        for run_dir in self.html_directory.iterdir()
                        if "cli_gui" in str(run_dir)]
            run_dirs.sort(reverse=True)

            last_run_dir = Path(run_dirs[0]) if len(run_dirs) > 0 else None
            if last_run_dir and not last_run_dir.joinpath('done').exists():
                # Use the old run directory
                logger.debug("Using HTML logging file from last time.")
                self.html_directory = self.html_directory.joinpath(
                    last_run_dir)
            else:
                # Create new run directory
                logger.debug("Creating new HTML logging file.")
                self.html_directory = self.html_directory.joinpath(
                    str(int(time())) + '_cli_gui')
        else:
            calling_func = inspect.stack()[1][3]
            self.html_directory = self.html_directory.joinpath(
                str(int(time())) + '_' + calling_func)

        self.screenshot_directory = self.html_directory.joinpath("screenshots")
        # will create both dirs
        self.screenshot_directory.mkdir(parents=True, exist_ok=True)

        self.html_file = self.html_directory.joinpath("index.html")
        if not self.html_file.exists():
            with open(self.html_file, 'w') as fp:
                fp.write(
                    "<html lang=\"en\">\n"
                    "<head>\n"
                    "<meta charset=\"UTF-8\">\n"
                    "<meta name=\"viewport\" content=\"width=device-width, "
                    "initial-scale=1.0\">\n"
                    "<title>Test Results</title>\n"
                    "</head>\n"
                    "<body style=\"background-color:#000;\">\n"
                )

        fmt = "<span style=\"color:limegreen;\">"
        fmt += "%(asctime)s</span> "
        fmt += "<span style=\"color:white;\">"
        fmt += "%(name)s:%(module)s.%(funcName)s.%(lineno)d </span>"
        fmt += "<span style=\"color:royalblue;\">[%(levelname)s] </span>"
        fmt += "<pre style=\"color:limegreen;\">%(message)s</pre>"
        self.fileHandler = logging.FileHandler(self.html_file)
        self.fileHandler.setLevel(logging.DEBUG)
        self.fileHandler.setFormatter(
            logging.Formatter("<p>" + fmt + "</p>")
        )

        if self.from_cli:
            logger.addHandler(self.fileHandler)

        self.mouse = Mouse()

        # workaround for keyboard library
        # otherwise the first character is not sent
        keyboard.send('enter')

        # create screen object to use from calls
        self.screen = Screen(self.screenshot_directory, self.html_file)

    def __enter__(self):
        # By restarting gdm, the system gets into defined state
        run(['systemctl', 'restart', 'gdm'], check=True)
        # Cannot screenshot before gdm starts displaying
        # This would break the display
        sleep(self.gdm_init_time)

        if not self.from_cli:
            logger.addHandler(self.fileHandler)

        return self

    def __exit__(self, type, value, traceback):
        done_file = self.html_directory.joinpath('done')
        print(done_file)
        if done_file.exists():
            return

        run(['systemctl', 'stop', 'gdm'], check=True)

        with open(self.html_file, 'a') as fp:
            fp.write(
                "</body>\n"
                "</html>\n"
            )

        print(done_file)
        with open(done_file, 'w') as fp:
            fp.write("done")

        if not self.from_cli:
            logger.removeHandler(self.fileHandler)

        logger.info(f"HTML file with results created in {self.html_directory}.")

    @action_decorator
    @log_decorator
    def click_on(self, key: str, timeout: float = 30):
        """Clicks matching word on the screen.

        :param key: String to find in the screenshot.
        :param timeout: If the key is not found within this timeout,
            an exception will be raised. Timeout is in seconds.
        """
        logger.info(f"Trying to find key='{key}' to click on.")

        end_time = time() + timeout
        item = None
        first_scr = None
        last_scr = None

        # Repeat screenshotting, until the key is found
        while time() < end_time:
            # Capture the screenshot
            screenshot = self.screen.screenshot()

            last_scr = screenshot
            if first_scr is None:
                first_scr = screenshot

            df = image_to_data(screenshot)
            selection = df['text'] == key

            # If there is no matching word, try again
            if selection.sum() == 0:
                logger.info('Found no match, trying again')
                continue

            # Exactly one word matching, exit the loop
            elif selection.sum() == 1:
                logger.info('Found exactly one match')
                item = df[selection].iloc[0]
                break

            # More than one word matches, choose the first match
            # Probably deterministic, but it should not be relied upon
            else:
                logger.info('Found multiple matches')
                item = df[selection].iloc[0]
                break

        if item is None:
            raise Exception(f"Found no key='{key}' in screenshots "
                            f"{first_scr} to {last_scr}")

        x = float(item['left'] + item['width'] / 2)
        y = float(item['top'] + item['height'] / 2)

        self.mouse.move(x, y)
        sleep(0.5)
        self.mouse.click()
        sleep(self.wait_time)

    @action_decorator
    @log_decorator
    def kb_write(self, *args, **kwargs):
        # delay is a workaround needed for keyboard library
        kwargs.setdefault('delay', 0.1)

        word = args[0]
        last = ""
        for char in word:
            if char.isupper():
                if last != "":
                    keyboard.write(*[last], **kwargs)
                    last = ""
                keyboard.send(f"shift+{char.lower()}")
            else:
                last = f"{last}{char}"
        keyboard.write(*[last], **kwargs)

    @action_decorator
    @log_decorator
    def kb_send(self, *args, **kwargs):
        keyboard.send(*args, **kwargs)

    @log_decorator
    def assert_text(self, key: str, timeout: float = 0):
        """
        Given key must be found in a screenshot before the timeout.

        If the key is not found, exception is raised.
        Zero timeout means that only one screenshot
        will be taken and evaluated.

        :param key: String to find in the screenshot.
        :param timeout: If the key is not found within this timeout,
            an exception will be raised. Timeout is in seconds.
        """

        logger.info(f"Trying to find key='{key}'")

        end_time = time() + timeout
        first = True

        while first or time() < end_time:
            first = False
            # Capture the screenshot
            screenshot = self.screen.screenshot()
            df = image_to_data(screenshot)
            selection = df['text'] == key

            # The key was found
            if selection.sum() != 0:
                return

        raise Exception('The key was not found.')

    @log_decorator
    def assert_no_text(self, key: str, timeout: float = 0):
        """
        If the given key is found in any screenshot before the timeout,
        an exception is raised.

        Zero timeout means that only one screenshot
        will be taken and evaluated.

        :param key: String that should not be found in the screenshot.
        :param timeout: Timeout is in seconds.
        """
        logger.info(f"Trying to find key='{key}'"
                    " (it should not be in the screenshot)")

        end_time = time() + timeout
        first = True

        while first or time() < end_time:
            first = False
            # Capture the screenshot
            screenshot = self.screen.screenshot()
            df = image_to_data(screenshot)
            selection = df['text'] == key

            # The key was found, but should not be
            if selection.sum() != 0:
                raise Exception(f"The key='{key}' was found "
                                f"in the screenshot {screenshot}")

    @log_decorator
    def check_home_screen(self, polarity: bool = True):
        """
        Check for the home screen to determine if user is logged in

        If OS version is defined as Fedora, we set the text for which to
        search to "tosearch" instead of the original "Activities". In later
        versions of Fedora, "Activities" text is no longer visible. "tosearch"
        should be visible on the login screen as the search bar is still
        present.

        After defining polarity and the string to check, run the appropriate
        function with the string to search for.

        :param polarity: Define whether to search for presence or absence of
            string indicating home screen is displayed.
        """
        if polarity is True:
            func_str = 'assert_text'
        else:
            func_str = 'assert_no_text'

        if isDistro('fedora'):
            check_str = 'tosearch'
        else:
            check_str = 'Activities'

        func = getattr(self, func_str)
        func(check_str, timeout=20)



"""
Exceptions that are used in the SCAutolib
"""


class SCAutolibException(Exception):
    """
    Base SCAutolib exception
    """
    def __init__(self, *args):
        super().__init__(*args)


class SCAutolibWrongConfig(SCAutolibException):
    default = "Key/section for current operation is not present in the " \
              "configuration file"

    def __init__(self, msg=None):
        msg = self.default if msg is None else msg
        super().__init__(msg)


class SCAutolibMissingCA(SCAutolibException):
    default = "CA is not configured on the system"

    def __init__(self, msg=None):
        msg = self.default if msg is None else msg
        super().__init__(msg)


class SCAutolibMissingUserConfig(SCAutolibException):
    def __init__(self, name):
        msg = f"User {name} is found in config file"
        super().__init__(msg)



"""
This module provides a set of additional helping functions that are used
across the library. These functions are based on library demands and are
not aimed to cover some general use-cases or specific corner cases.
"""
import json
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from pathlib import Path

from SCAutolib import (run, logger, TEMPLATES_DIR, LIB_DUMP_USERS, LIB_DUMP_CAS,
                       LIB_DUMP_CARDS)
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.models.CA import LocalCA, BaseCA, CustomCA, IPAServerCA
from SCAutolib.models.card import Card
from SCAutolib.models.file import OpensslCnf, SSSDConf
from SCAutolib.models.user import User


def _check_selinux():
    """
    Checks if specific SELinux module for virtual smart card is installed.
    This is implemented be checking the hardcoded name for the module
    (virtcacard) to be present in the list of SELinux modules. If this name is
    not present in the list, then virtcacard.cil file would be created in conf
    or subdirectory in the CA directory specified by the configuration file.
    """
    result = run("semodule -l", print_=False)
    if "virtcacard" not in result.stdout:
        logger.debug(
            "SELinux module for virtual smart cards is not present in the "
            "system. Installing...")

        run(["semodule", "-i", f"{TEMPLATES_DIR}/virtcacard.cil"])

        run(["systemctl", "restart", "pcscd"])
        logger.debug("pcscd service is restarted")

    logger.debug(
        "SELinux module for virtual smart cards is installed")


def _gen_private_key(key_path: Path):
    """
    Generate RSA private key to specified location.

    :param key_path: path to output certificate
    """
    # CAC specification do not specify key size specifies key size
    # up to 2048 bits, so keys greater than 2048 bits is not supported
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    with key_path.open("wb") as f:
        f.write(key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()))


def _install_packages(packages):
    """
    Install given packages and log package version

    :param packages: list of packages to be installed
    """
    run(f"dnf install -y {' '.join(packages)}")
    for pkg in packages:
        pkg = run(["rpm", "-q", pkg]).stdout
        logger.debug(f"Package {pkg} is installed")


def _check_packages(packages):
    """
    Find missing packages

    :param packages: list of required packages
    :type packages: list
    :return: list of missing packages
    """
    missing = []
    for pkg in packages:
        # Return code 1 means the package is not installed
        out = run(["rpm", "-q", pkg], return_code=[0, 1])
        if out.returncode == 1:
            logger.warning(f"Package {pkg} is required for the testing, "
                           f"but is not present in the system")
            missing.append(pkg)
        else:
            logger.debug(f"Package {out.stdout.strip()} is present")
    return missing


def dump_to_json(obj):
    """
    Store serialised object to the JSON file.
    """
    with obj.dump_file.open("w") as f:
        json.dump(obj.to_dict(), f)
    logger.debug(f"Object {type(obj)} is stored to the {obj.dump_file} file")


def load_user(username, **kwargs):
    """
    Load user with given username from JSON file.

    :param username: username of the user
    :type username: str

    :return: user object
    :rtype: BaseUser
    """
    user_file = LIB_DUMP_USERS.joinpath(f"{username}.json")
    logger.debug(f"Loading user {username} from {user_file}")
    user = None
    if user_file.exists():
        user = User.load(user_file, **kwargs)
    else:
        raise SCAutolibException(f"{user_file} does not exist")
    return user


def load_token(card_name: str = None, update_sssd: bool = False):
    """
    Load card with given name from JSON file. This function is intended to load
    card objects to tests during pytest configuration. If update_sssd param is
    True sssd.conf file will be updated based on card data

    :param card_name: name of the card to be loaded
    :type card_name: str
    :param update_sssd: indicates if sssd.conf matchrule should be updated based
        on card data
    :type update_sssd bool

    :return: card object
    :rtype: Card
    """
    card_file = LIB_DUMP_CARDS.joinpath(f"{card_name}.json")
    logger.debug(f"Loading card {card_name} from {card_file}")
    card = None
    if card_file.exists():
        card = Card.load(card_file)
    if update_sssd:
        sssd_conf = SSSDConf()
        sssd_conf.set(section=f"certmap/shadowutils/{card.cardholder}",
                      key="matchrule",
                      value=f"<SUBJECT>.*CN={card.CN}.*")
        sssd_conf.save()
        run(["sss_cache", "-E"])
        run(["systemctl", "restart", "sssd"])
    return card


def ipa_factory():
    """
    Create a new IPAServerCA object.

    .. note: Creating new IPA server with CA is not supported.

    :return: object of IPAServerCA
    :rtype: SCAutolib.models.CA.IPAServerCA
    """
    json_file = LIB_DUMP_CAS.joinpath("ipa-server.json")
    if not json_file.exists():
        msg = "Dump file for ipa server CA is not present."
        logger.error(msg)
        logger.error("The reason for this is most likely that the system was "
                     "not configured for IPA client via SCAutolib")
        raise SCAutolibException(msg)
    ca = BaseCA.load(json_file)
    if not isinstance(ca, IPAServerCA):
        msg = "Values in dump file are not valid for IPA server, so the " \
              "object can't be created"
        logger.error(msg)
        raise SCAutolibException(msg)
    return ca


def ca_factory(path: Path = None, cnf: OpensslCnf = None,
               card_data: dict = None, ca_name: str = None,
               create: bool = False):
    """
    Create CA object. If certain CA object was created in previous run of
    SCAutolib and it was serialized and saved in .json file, then such CA object
    would be initialized based on the file. If create param is True new CA
    object will be created regardless the presence of the .json file.

    :param path: path to the CA directory
    :type path: Path
    :param cnf: object representing openssl cnf file
    :type cnf: OpensslCnf object
    :param card_data: dictionary with various attributes of the card as PIN,
        cardholder, slot, etc.
    :type card_data: dict
    :param ca_name: name of CA that identifies CA file to be loaded if create
        parameter is set to False
    :type ca_name: str
    :param create: indicator to create new CA. If it's false existing CA files
        will be loaded
    :type create: bool
    :return: CA object
    :rtype: SCAutolib.models.CA object
    """
    if not create:
        ca = BaseCA.load(LIB_DUMP_CAS.joinpath(f"{ca_name}.json"))
        return ca

    if not path:            # create CA for physical card
        ca = CustomCA(card_data)
        return ca
    else:                   # create new CA object for virtual card
        ca = LocalCA(root_dir=path, cnf=cnf)
        return ca



"""
This module contains unit tests for OpensslCnf
"""
from pathlib import Path

from SCAutolib.models.file import OpensslCnf
from conftest import FILES_DIR
from test_sssd_conf import compare_conf_files


def test_create(tmpdir):
    reference_file = Path(FILES_DIR, "openssl_cnf_test_result")
    tmpfile = Path(tmpdir, "testfile")
    opensslcnf_test = OpensslCnf(filepath=tmpfile, conf_type="user",
                                 replace=["test_user_123", "test_cn"])
    opensslcnf_test.create()
    opensslcnf_test.save()
    compare_conf_files(opensslcnf_test._conf_file, reference_file)
